#include <stdio.h>
#include "defs.h"
#include "dep.h"
#include "gram.h"
#include "new.h"
#include "state.h"

LOCAL unsigned **gsets;
LOCAL short *smap;

build_gsets()
{
  register int i, sym, stateno, target;
  register short *asym, *to_state;
  register shifts *sp;
  register core *stp;
  register int ssetsize;

  ssetsize = WORDSIZE(nstates);

  gsets = NEW2(nvars, unsigned *);
  for (i = nvars - 1; i >= 0; i--)
    gsets[i] = NEW2(ssetsize, unsigned);
  gsets -= start_symbol;

  asym = NEW2(nstates, short);
  for (stp = first_state; stp; stp = stp->next)
    asym[stp->number] = stp->accessing_symbol;

  for (sp = first_shift; sp; sp = sp->next)
    {
      stateno = sp->number;
      to_state = sp->shifts;
      for (i = sp->nshifts - 1; i >= 0; i--)
	{
	  target = to_state[i];
	  sym = asym[target];
	  if (ISVAR(sym))
	    SETBIT(gsets[sym], stateno);
	  else
	    break;
	}
    }
  FREE(asym);
}

collapse_gsets()
{
  register int i, j, k;
  register int done, common, setsize;
  register unsigned *set, *sset;

  gsets += start_symbol;
  setsize = WORDSIZE(nstates);

  done = 0;
  while ( ! done)
    {
      done = 1;
      for (i = 1; i < nvars; i++)
	{
	  set = gsets[i];
	  if (set)
	    {
	      common = 0;
	      for (j = 0; j < i && ! common; j++)
		{
		  sset = gsets[j];
		  if (sset)
		    {
		      for (k = 0; k < setsize && ! common; k++)
			{
			  if (set[k] & sset[k])
			    common = 1;
			}
		      if (common)
			{
			  done = 0;
			  for (k = 0; k < setsize; k++)
			    sset[k] |= set[k];
			  FREE(set);
			  gsets[i] = NIL(unsigned);
			}
		    }
		}
	    }
	}
    }
}

build_smap()
{
  register int i, j, k;
  register unsigned *set;

  smap = NEW2(nstates, short);

  k = 0;
  for (i = 0; i < nvars; i++)
    {
      set = gsets[i];
      if (set)
	{
	  for (j = 0; j < nstates; j++)
	    if (BIT(set, j))
	      {
		smap[j] = k;
		k++;
	      }
	}
    }

  for (i = 1; i < nstates; i++)
    {
      if (smap[i] == 0)
	{
	  smap[i] = k;
	  k++;
	}
    }

  for (i = 0; i < nvars; i++)
    FREE(gsets[i]);
  FREE(gsets);
}

renumber()
{
  register int i;
  register short *to_state;
  register core *stp;
  register shifts *sp;
  register reductions *rp;

  for (stp = first_state; stp; stp = stp->next)
    stp->number = smap[stp->number];

  for (sp = first_shift; sp; sp = sp->next)
    {
      sp->number = smap[sp->number];
      to_state = sp->shifts;
      for (i = sp->nshifts - 1; i >= 0; i--)
	to_state[i] = smap[to_state[i]];
    }

  for (rp = first_reduction; rp; rp = rp->next)
    rp->number = smap[rp->number];

  FREE(smap);
}

optim()
{
  build_gsets();
  collapse_gsets();
  build_smap();
  renumber();
}
